/**
 * A Monarch language definition
 */
export interface IMonarchLanguage {
    /**
     * map from string to ILanguageRule[]
     */
    tokenizer: {
        [name: string]: IMonarchLanguageRule[];
    };
    /**
     * is the language case insensitive?
     */
    ignoreCase?: boolean;
    /**
     * is the language unicode-aware? (i.e., /\u{1D306}/)
     */
    unicode?: boolean;
    /**
     * if no match in the tokenizer assign this token class (default 'source')
     */
    defaultToken?: string;
    /**
     * for example [['{','}','delimiter.curly']]
     */
    brackets?: IMonarchLanguageBracket[];
    /**
     * start symbol in the tokenizer (by default the first entry is used)
     */
    start?: string;
    [attr: string]: any;
}
/**
 * A rule is either a regular expression and an action
 * 		shorthands: [reg,act] == { regex: reg, action: act}
 *		and       : [reg,act,nxt] == { regex: reg, action: act{ next: nxt }}
 */
export declare type IShortMonarchLanguageRule1 = [string | RegExp, IMonarchLanguageAction];
export declare type IShortMonarchLanguageRule2 = [string | RegExp, IMonarchLanguageAction, string];
export interface IExpandedMonarchLanguageRule {
    /**
     * match tokens
     */
    regex?: string | RegExp;
    /**
     * action to take on match
     */
    action?: IMonarchLanguageAction;
    /**
     * or an include rule. include all rules from the included state
     */
    include?: string;
}
export declare type IMonarchLanguageRule = IShortMonarchLanguageRule1 | IShortMonarchLanguageRule2 | IExpandedMonarchLanguageRule;
/**
 * An action is either an array of actions...
 * ... or a case statement with guards...
 * ... or a basic action with a token value.
 */
export declare type IShortMonarchLanguageAction = string;
export interface IExpandedMonarchLanguageAction {
    /** array of actions for each parenthesized match group */
    group?: IMonarchLanguageAction[];
    /** map from string to ILanguageAction */
    cases?: Object;
    /** token class (ie. css class) (or "@brackets" or "@rematch") */
    token?: string;
    /** the next state to push, or "@push", "@pop", "@popall" */
    next?: string;
    opens?: string;
    closes?: string;
    /** switch to this state */
    switchTo?: string;
    /** go back n characters in the stream */
    goBack?: number;
    /** @open or @close */
    bracket?: string;
    /** switch to embedded language (using the mimetype) or get out using "@pop" */
    nextEmbedded?: string;
    /** log a message to the browser console window */
    log?: string;
}
export declare type IMonarchLanguageAction = IShortMonarchLanguageAction | IExpandedMonarchLanguageAction | (IShortMonarchLanguageAction | IExpandedMonarchLanguageAction)[];
/**
 * This interface can be shortened as an array, ie. ['{','}','delimiter.curly']
 */
export interface IMonarchLanguageBracket {
    /**
     * open bracket
     */
    open: string;
    /**
     * closing bracket
     */
    close: string;
    /**
     * token class
     */
    token: string;
}
export declare const enum MonarchBracket {
    None = 0,
    Open = 1,
    Close = -1
}
export interface ILexerMin {
    ignoreCase: boolean;
    unicode: boolean;
    defaultToken: string;
    stateNames: {
        [stateName: string]: any;
    };
    [attr: string]: any;
}
export interface ILexer extends ILexerMin {
    maxStack: number;
    start: string | null;
    ignoreCase: boolean;
    unicode: boolean;
    tokenTypes: Set<string>;
    tokenizer: {
        [stateName: string]: IRule[];
    };
    brackets: IBracket[];
}
export interface IBracket {
    token: string;
    open: string;
    close: string;
}
export declare type FuzzyAction = IAction | string;
export declare function isFuzzyActionArr(what: FuzzyAction | FuzzyAction[]): what is FuzzyAction[];
export declare function isFuzzyAction(what: FuzzyAction | FuzzyAction[]): what is FuzzyAction;
export declare function isString(what: FuzzyAction): what is string;
export declare function isIAction(what: FuzzyAction): what is IAction;
export interface IRule {
    regex: RegExp;
    action: FuzzyAction;
    matchOnlyAtLineStart: boolean;
    name: string;
}
export interface IAction {
    group?: FuzzyAction[];
    test?: (id: string, matches: string[], state: string, eos: boolean) => FuzzyAction;
    token?: string;
    tokenSubst?: boolean;
    next?: string;
    nextEmbedded?: string;
    bracket?: MonarchBracket;
    log?: string;
    switchTo?: string;
    goBack?: number;
    transform?: (states: string[]) => string[];
    opens?: string;
    closes?: string;
}
export interface IBranch {
    name: string;
    value: FuzzyAction;
    test?: (id: string, matches: string[], state: string, eos: boolean) => boolean;
}
/**
 * Is a string null, undefined, or empty?
 */
export declare function empty(s: string): boolean;
/**
 * Puts a string to lower case if 'ignoreCase' is set.
 */
export declare function fixCase(lexer: ILexerMin, str: string): string;
/**
 * Ensures there are no bad characters in a CSS token class.
 */
export declare function sanitize(s: string): string;
/**
 * Logs a message.
 */
export declare function log(lexer: ILexerMin, msg: string): void;
export declare function createError(lexer: ILexerMin, msg: string): Error;
/**
 * substituteMatches is used on lexer strings and can substitutes predefined patterns:
 * 		$$  => $
 * 		$#  => id
 * 		$n  => matched entry n
 * 		@attr => contents of lexer[attr]
 *
 * See documentation for more info
 */
export declare function substituteMatches(lexer: ILexerMin, str: string, id: string, matches: string[], state: string): string;
/**
 * Find the tokenizer rules for a specific state (i.e. next action)
 */
export declare function findRules(lexer: ILexer, inState: string): IRule[] | null;
/**
 * Is a certain state defined? In contrast to 'findRules' this works on a ILexerMin.
 * This is used during compilation where we may know the defined states
 * but not yet whether the corresponding rules are correct.
 */
export declare function stateExists(lexer: ILexerMin, inState: string): boolean;
